use actix_web::{
    error::InternalError, get, middleware::Logger, post, web, App, HttpResponse, HttpServer,
    Responder,
};
use serde::{Deserialize, Serialize};
use serde_json::json;
use std::sync::Arc;
use verax::{AccountId, AnyId, AssetDefinition, AssetManager, Status, TransactionId, Type};

#[derive(Deserialize)]
pub struct Movement {
    pub account: AccountId,
    pub amount: String,
    pub asset: String,
}

#[derive(Deserialize)]
pub struct Deposit {
    pub account: AccountId,
    pub amount: String,
    pub asset: String,
    pub memo: String,
    pub status: Status,
}

impl Deposit {
    pub async fn to_ledger_transaction(
        self,
        ledger: &Ledger,
    ) -> Result<verax::Transaction, verax::Error> {
        let amount = ledger._inner.parse_amount(&self.asset, &self.amount)?;
        ledger
            ._inner
            .deposit(&self.account, amount, self.status, self.memo)
            .await
    }
}

#[derive(Deserialize)]
pub struct Transaction {
    pub debit: Vec<Movement>,
    pub credit: Vec<Movement>,
    pub memo: String,
    pub status: Status,
}

#[derive(Deserialize)]
pub struct UpdateTransaction {
    pub status: Status,
    pub memo: String,
}

impl UpdateTransaction {
    pub async fn to_ledger_transaction(
        self,
        id: &TransactionId,
        ledger: &Ledger,
    ) -> Result<verax::Transaction, verax::Error> {
        ledger
            ._inner
            .change_status(id, self.status, self.memo)
            .await
    }
}

impl Transaction {
    pub async fn to_ledger_transaction(
        self,
        ledger: &Ledger,
    ) -> Result<verax::Transaction, verax::Error> {
        let from = self
            .debit
            .into_iter()
            .map(|x| {
                ledger
                    ._inner
                    .parse_amount(&x.asset, &x.amount)
                    .map(|amount| (x.account, amount))
            })
            .collect::<Result<Vec<_>, _>>()?;

        let to = self
            .credit
            .into_iter()
            .map(|x| {
                ledger
                    ._inner
                    .parse_amount(&x.asset, &x.amount)
                    .map(|amount| (x.account, amount))
            })
            .collect::<Result<Vec<_>, _>>()?;

        ledger
            ._inner
            .new_transaction(self.memo, self.status, from, to)
            .await
    }
}

#[derive(Serialize, Deserialize)]
struct Item {
    id: i32,
    name: String,
}

#[derive(Serialize)]
struct AccountResponse {
    amount: String,
    asset: Arc<str>,
}

#[get("/balance/{id}")]
async fn get_balance(info: web::Path<AccountId>, ledger: web::Data<Ledger>) -> impl Responder {
    let asset_manager = ledger._inner.asset_manager();
    match ledger._inner.get_balance(&info.0).await {
        Ok(balances) => {
            match balances
                .into_iter()
                .map(|amount| {
                    asset_manager
                        .get_name(amount.asset().id)
                        .map(|asset| AccountResponse {
                            amount: amount.to_string(),
                            asset,
                        })
                })
                .collect::<Result<Vec<_>, _>>()
            {
                Ok(balances) => HttpResponse::Ok().json(balances),
                Err(err) => HttpResponse::InternalServerError()
                    .json(json!({ "text": err.to_string(), "err": err})),
            }
        }
        Err(err) => HttpResponse::BadRequest().json(json!({ "text": err.to_string(), "err": err})),
    }
}

#[get("/{id}")]
async fn get_info(info: web::Path<AnyId>, ledger: web::Data<Ledger>) -> impl Responder {
    let result = match info.0 {
        AnyId::Account(account_id) => ledger
            ._inner
            .get_transactions(
                &account_id,
                vec![
                    Type::Deposit,
                    Type::Withdrawal,
                    Type::Exchange,
                    Type::Transaction,
                ],
            )
            .await
            .map(|transactions| HttpResponse::Ok().json(transactions)),
        AnyId::Transaction(transaction_id) => ledger
            ._inner
            .get_transaction(&transaction_id)
            .await
            .map(|tx| HttpResponse::Ok().json(tx)),
        AnyId::Payment(payment_id) => ledger
            ._inner
            .get_payment_info(&payment_id)
            .await
            .map(|tx| HttpResponse::Ok().json(tx)),
    };

    match result {
        Ok(x) => x,
        Err(err) => {
            HttpResponse::InternalServerError().json(json!({ "text": err.to_string(), "err": err}))
        }
    }
}

#[post("/deposit")]
async fn deposit(item: web::Json<Deposit>, ledger: web::Data<Ledger>) -> impl Responder {
    // Insert the item into a database or another data source.
    // For this example, we'll just echo the received item.
    match item.into_inner().to_ledger_transaction(&ledger).await {
        Ok(tx) => {
            // Insert the item into a database or another data source.
            // For this example, we'll just echo the received item.
            HttpResponse::Created().json(tx)
        }
        Err(err) => HttpResponse::BadRequest().json(json!({ "text": err.to_string(), "err": err})),
    }
}

#[post("/tx")]
async fn create_transaction(
    item: web::Json<Transaction>,
    ledger: web::Data<Ledger>,
) -> impl Responder {
    match item.into_inner().to_ledger_transaction(&ledger).await {
        Ok(tx) => HttpResponse::Accepted().json(tx),
        Err(err) => {
            HttpResponse::InternalServerError().json(json!({ "text": err.to_string(), "err": err}))
        }
    }
}

#[post("/{id}")]
async fn update_status(
    info: web::Path<TransactionId>,
    item: web::Json<UpdateTransaction>,
    ledger: web::Data<Ledger>,
) -> impl Responder {
    match item
        .into_inner()
        .to_ledger_transaction(&info.0, &ledger)
        .await
    {
        Ok(tx) => HttpResponse::Accepted().json(tx),
        Err(err) => {
            HttpResponse::InternalServerError().json(json!({ "text": err.to_string(), "err": err}))
        }
    }
}

pub struct Ledger {
    _inner: verax::Ledger<'static, verax::Batch<'static>, verax::SQLite<'static>>,
}

#[actix_web::main]
async fn main() -> std::io::Result<()> {
    if std::env::var_os("RUST_LOG").is_none() {
        std::env::set_var("RUST_LOG", "actix_web=info");
    }
    env_logger::init_from_env(env_logger::Env::new().default_filter_or("info"));

    let asset_manager = AssetManager::new(vec![
        AssetDefinition::new(1, "BTC", 8),
        AssetDefinition::new(2, "USD", 4),
    ]);

    let settings = "sqlite:./test.db"
        .parse::<verax::sqlite::SqliteConnectOptions>()
        .expect("valid settings")
        .create_if_missing(true);

    let pool = verax::sqlite::SqlitePoolOptions::new()
        .connect_with(settings)
        .await
        .expect("pool");

    let storage = verax::SQLite::new(pool.clone(), asset_manager.clone());
    storage.setup().await.expect("setup");

    HttpServer::new(move || {
        let storage = verax::SQLite::new(pool.clone(), asset_manager.clone());
        let ledger = verax::Ledger::new(storage, asset_manager.clone());
        App::new()
            .wrap(Logger::default())
            .app_data(web::Data::new(Ledger { _inner: ledger }))
            .app_data(web::JsonConfig::default().error_handler(|err, _req| {
                InternalError::from_response(
                    "",
                    HttpResponse::BadRequest()
                        .content_type("application/json")
                        .body(format!(r#"{{"error":"{}"}}"#, err)),
                )
                .into()
            }))
            .service(get_balance)
            .service(get_info)
            .service(deposit)
            .service(create_transaction)
            .service(update_status)
    })
    .bind("127.0.0.1:8080")?
    .run()
    .await
}
