use serde::{Deserialize, Serialize};

#[derive(thiserror::Error, Debug)]
pub enum Error {
    #[error("Invalid status: {0}")]
    InvalidStatus(u32),
}

#[derive(Debug, Clone, Copy, PartialEq, Serialize, Deserialize)]
#[serde(rename_all = "snake_case")]
/// Transaction types
pub enum Type {
    /// Deposit
    Deposit,
    /// Withdrawal
    Withdrawal,
    /// Transaction
    Transaction,
    /// Internal / Exchange Transaction
    Exchange,
}

impl Type {
    #[inline]
    /// Is it a transaction
    ///
    /// Transactions have a constraint that the amounts and assets must be equal
    /// on both sides. Withdrawals and deposits don't have this constraint.
    pub fn is_transaction(&self) -> bool {
        matches!(self, Self::Transaction | Self::Exchange)
    }
}

impl TryFrom<u32> for Type {
    type Error = Error;
    fn try_from(value: u32) -> Result<Self, Self::Error> {
        match value {
            0 => Ok(Self::Transaction),
            1 => Ok(Self::Deposit),
            2 => Ok(Self::Withdrawal),
            1000 => Ok(Self::Exchange),
            _ => Err(Error::InvalidStatus(value)),
        }
    }
}

impl From<Type> for u32 {
    fn from(value: Type) -> Self {
        (&value).into()
    }
}

impl From<&Type> for u32 {
    fn from(value: &Type) -> Self {
        match value {
            Type::Transaction => 0,
            Type::Deposit => 1,
            Type::Withdrawal => 2,
            Type::Exchange => 1000,
        }
    }
}
